/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef __VIPS_MARSHAL_H__
#define __VIPS_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* INT: VOID (vipsmarshal.list:25) */
extern
void vips_INT__VOID (GClosure     *closure,
                     GValue       *return_value,
                     guint         n_param_values,
                     const GValue *param_values,
                     gpointer      invocation_hint,
                     gpointer      marshal_data);

/* INT64: INT64, INT (vipsmarshal.list:26) */
extern
void vips_INT64__INT64_INT (GClosure     *closure,
                            GValue       *return_value,
                            guint         n_param_values,
                            const GValue *param_values,
                            gpointer      invocation_hint,
                            gpointer      marshal_data);

/* INT64: POINTER, INT64 (vipsmarshal.list:27) */
extern
void vips_INT64__POINTER_INT64 (GClosure     *closure,
                                GValue       *return_value,
                                guint         n_param_values,
                                const GValue *param_values,
                                gpointer      invocation_hint,
                                gpointer      marshal_data);


G_END_DECLS

#endif /* __VIPS_MARSHAL_H__ */
